# pmmail_folder_check.pl
#
# PMMail folder checker to check consistency of the PMMail mail store 
#
# Copyright Dave Saville 2011 2012
# dave@deezee.org
# All rights reserved
#
# If it breaks you get to keep *all* the pieces :-)

use strict;
use warnings;
use Carp;

#use Data::Dumper;

print "PMMail_folder_check Version 2.6.2\n";

my $pmmail_dir = shift;
my $max_errors = shift || 0;

unless ( defined $pmmail_dir )
{
  print(
    "PMMail_folder_check takes two arguments\n",
    "The first is the location of your PMMail mailstore\n",
    "ie the directory holding the *.ACT directories\n",
    "The second (optional) is the maximum mail missing messages per folder\n",
    "The script will loop round all *.ACT directories found\n",
    "\n",
    "Logging is to STDOUT\n",
    "It will log message files that are not in folder.bag or folder.xml files\n",
    "It will log the converse\n",
    "It will log zero length bag files if the folder contains messages\n",
    "It will log zero byte folder.ini files\n",
    "It will log duplicate folder ids\n",
    "It will log archiving discrepancies\n",
    "Note summary sizes are only the sizes of *.MSG files\n",
    "Quick 'n dirty validation of XML files\n",
    "Validates folder.uct\n",
  );
  exit;
} ## end unless ( defined $pmmail_dir...)

print "WARNING: PMMail is running - Results may be inconsistent\n"
  if grep /PMMAIL.EXE/, `pstat /C`;  # check if running
print "\nMessage warnings limited to $max_errors\n" if $max_errors;
$pmmail_dir =~ s{\\}{/}g;            # \ to /
$pmmail_dir .= '/' if $pmmail_dir !~ m{/$};  # ensure trailing /
my $total_size = 0;
my @acts;                                    # array of accounts
my @logs;

# get all accounts

opendir my $PMMAIL, $pmmail_dir or croak "Can't open $pmmail_dir $!";

foreach my $member ( readdir $PMMAIL )
{
  push @acts, $member if $member =~ m{\.ACT$}i;
  push @logs, $member if $member =~ m{^logfile}i;
}

closedir $PMMAIL;

unless (@acts)
{
  print "No *.ACT folders found\n",
    "Point at the directory holding your *.ACT directories\n";
  exit 1;
} ## end unless (@acts)

my $pmmail_version = q{};  # empty string

# try and find the PMMail version from the logfiles
# we read all of each log as we want the last logged version

if ( open my $LOG, '<', "$pmmail_dir/logfile.log" )
{
  while ( my $line = <$LOG> )
  {
    if ( $line =~ m{(\([^\)]+\)) Built on} )
    {
      $pmmail_version = $1;
    }
  } ## end while ( my $line = <$LOG>...)

  close $LOG;
} ## end if ( open my $LOG, '<'...)

unless ($pmmail_version)
{
LOG:
  foreach my $log ( reverse @logs )
  {
    open my $LOG, '<', "$pmmail_dir/$log" or croak "Can't open $log $!";

    while ( my $line = <$LOG> )
    {
      if ( $line =~ m{(\([^\)]+\)) Built on} )
      {
        $pmmail_version = $1;
      }
    } ## end while ( my $line = <$LOG>...)

    close $LOG;
    last LOG if $pmmail_version;
  } ## end foreach my $log ( reverse @logs...)
} ## end unless ($pmmail_version)

print "PMMAIL Version (from log files) $pmmail_version\n";
our %folder_name_from;
my $folder_count;
my $folder_msg_count;
my $folder_next_id;
my $account_error_count = 0;
my $total_error_count   = 0;
my $act_size            = 0;

foreach my $act (@acts)
{
  $folder_count        = 0;
  $folder_msg_count    = 0;
  $folder_next_id      = 0;
  $account_error_count = 0;
  %folder_name_from    = ();
  my %XML_file;
  my $ini_account_name = '?';

  opendir my $ACT, $pmmail_dir . $act
    or croak "Can't open $pmmail_dir$act $!";

  foreach my $member ( readdir $ACT )
  {
    $XML_file{$member} = 1 if $member =~ m{\.xml}i;
  }

  close $ACT;

  if ( open my $INI, '<', $pmmail_dir . $act . '/acct.ini' )
  {
    binmode $INI;
    local $/ = undef;
    my $ini = <$INI>;
    close $INI;

    if ( $ini =~ m{ACCTNAME\x00+([^\x00]+)}s )
    {
      $ini_account_name = $1;
    }
  } ## end if ( open my $INI, '<'...)
  else
  {
    print "WARNING: Can't open $pmmail_dir$act/acct.ini $!";
    $account_error_count++;
  }

  if ( is_valid_XML( $pmmail_dir . $act . '/acctpr.xml', \%XML_file ) )
  {
    if ( open my $XML, '<', $pmmail_dir . $act . '/acctpr.xml' )
    {
      local $/ = undef;
      my $xml = <$XML>;
      close $XML;
      my $XML_account_name;

      if ( $xml =~ m{>([^<]+)</profile>}s )
      {
        $XML_account_name = $1;
      }

      my $iname = $ini_account_name;
      my $xname = $XML_account_name;

      # names in XML have HTML style &nnn; special chars
      # so we take them out replacing them and all other
      # special characters with space
      # do the same with ini name so we can get a compare to match

      $iname =~ s{[^a-z0-9]}{ }ig;
      $xname =~ s{&[^;]+;}{ }g;
      $xname =~ s{[^a-z0-9]}{ }ig;

      if ( $iname ne $xname )
      {
        print "\nWARNING: Account names differ in acct.ini and acctpr.xml\n";
        print "\t>$ini_account_name< >$XML_account_name<\n";
        $account_error_count++;
      } ## end if ( $iname ne $xname ...)
    } ## end if ( open my $XML, '<'...)
    else
    {
      print "WARNING: Can't open $pmmail_dir$act/acctpr.xml $!";
      $account_error_count++;
    }

  } ## end if ( is_valid_XML( $pmmail_dir...))

  print "\nProcessing account: $ini_account_name ($pmmail_dir$act)\n";
  $act_size = 0;
  folder( $pmmail_dir . $act, $ini_account_name );
  printf
    "\n\tSize: %s; Folders: %s; eMails: %s; Next Folder ID: %s; Errors: %s\n",
    commify($act_size), commify($folder_count), commify($folder_msg_count),
    commify( $folder_next_id + 1 ), commify($account_error_count);
  $total_error_count += $account_error_count;
  $total_size        += $act_size;
} ## end foreach my $act (@acts)

printf "\nProcessing completed: Size: %s; Errors: %s\n", commify($total_size),
  commify($total_error_count);
exit( ( $total_error_count < 255 ) ? $total_error_count : 255 );

sub folder
{
  my $this_folder         = shift;
  my $this_name           = shift;
  my $folder_error_count  = 0;
  my $message_error_count = 0;
  my $has_XML             = 0;
  my $PR_XML              = 1;
  my $IX_XML              = 2;
  my $ini_folder_name     = q{};   # empty string
  my $folder_name;
  my $id              = 0;
  my $bag_count       = 0;
  my $bag_unread      = 0;
  my $uct_count       = 0;
  my $uct_unread      = 0;
  my $XML_pr_count    = 0;
  my $XML_pr_unread   = 0;
  my $XML_ix_count    = 0;
  my $XML_ix_unread   = 0;
  my $warned_file     = 0;
  my $warned_name     = 0;
  my $is_archived     = 0;
  my $ini_is_archived = 0;
  my $XML_is_archived = 0;

  my @folders;

  my %admin_file;
  my %bag_msg;
  my %folder_msg;
  my %XML_msg;

  $folder_count++;
  opendir my $DIR, $this_folder
    or croak "Can't open $this_folder\n($this_name/$folder_name)\n$!";

  # build directory caches

  foreach my $member ( readdir $DIR )
  {
    my ( $size, $mtime )
      = ( ( stat $this_folder . '/' . $member )[ 7, 9 ] );  # size and mtime

    if ( -f _ )
    {
      my ( $name, $suffix, $gz ) = split /\./, $member;

      if ( $suffix && $suffix =~ m{^msg$}i )
      {
        $act_size += $size;
        $folder_msg{ lc $name } = [ $size, $mtime, $member, $gz ];
      }
      else
      {
        $admin_file{$member} = [ $size, $mtime ];
      }
    } ## end if ( -f _ )
    elsif ( -d _ )
    {
      if ( $member =~ m{\.fld$}i )
      {
        push @folders, $member;
      }
    } ## end elsif ( -d _ )
  } ## end foreach my $member ( readdir...)

  close $DIR;

  if ( $this_folder =~ m{\.fld$}i )
  {
    if ( exists $admin_file{'folder.ini'}
         && $admin_file{'folder.ini'}[0] == 0 )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name, q{},
                             \$warned_name );
      print "\tfolder.ini is zero bytes\n";
      $folder_error_count++;
    } ## end if ( exists $admin_file...)
    else
    {
      if ( open my $INI, '<', "$this_folder/folder.ini" )
      {
        ( $ini_folder_name, $id, $ini_is_archived )
          = ( split /\xDE/, <$INI> )[ 0, 1, 10 ];
        close $INI;
      } ## end if ( open my $INI, '<'...)
      else
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name, q{},
                               \$warned_name );
        print "\tCan't open $this_folder/folder.ini $!\n";
        $folder_error_count++;
      } ## end else [ if ( open my $INI, '<'...)]
    } ## end else [ if ( exists $admin_file...)]

    if ( is_valid_XML( "$this_folder/folderpr.xml", \%admin_file ) )
    {
      $has_XML |= $PR_XML;

      if ( open my $XML, '<', "$this_folder/folderpr.xml" )
      {
        local $/ = undef;
        my $xml = <$XML>;
        close $XML;
        my $XML_folder_name;

        if ( $xml =~ m{>([^<]+)</profile>}s )
        {
          $XML_folder_name = $1;
        }

        if ( $xml =~ m{is-archived="(\d+)"}s )
        {
          $XML_is_archived = $1;
        }

        my $iname = $ini_folder_name;
        my $xname = $XML_folder_name;

        # clobber HTMLish entities

        $xname =~ s{&[^;]+;}{ }g;
        $xname =~ s{[^a-z0-9]}{ }ig;
        $iname =~ s{[^a-z0-9]}{ }ig;

        if ( $iname ne $xname )
        {
          print_warning_headers( $this_folder, \$warned_file, $this_name,
                                 q{}, \$warned_name );
          print "\tFolder names differ in folder.ini and folderpr.xml\n";
          print "\t>$ini_folder_name< >$XML_folder_name<\n";
          $folder_error_count++;
        } ## end if ( $iname ne $xname ...)

        if ( $xml =~ m{folder-id="([^"]+)}s )
        {
          $id = $1;
        }

        ( $XML_pr_count, $XML_pr_unread )
          = ( $xml =~ m{<message-counts total="(\d+)" unread="(\d+)"} );
      } ## end if ( open my $XML, '<'...)
      else
      {
        warning( $this_folder, \$warned_file, $this_name, q{},
                 \$warned_name );
        print "\tCan't open $this_folder/folderpr.xml $!\n";
        $folder_error_count++;
      } ## end else [ if ( open my $XML, '<'...)]
    } ## end if ( is_valid_XML( "$this_folder/folderpr.xml"...))

    $is_archived = $ini_is_archived & $XML_is_archived;
    $folder_name = $ini_folder_name;

    if ( $id > $folder_next_id && $id < 32000 )
    {
      $folder_next_id = $id;
    }

    if ( $ini_is_archived && !$XML_is_archived )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tFolder ini says archived and XML says not\n";
      $folder_error_count++;
    } ## end if ( $ini_is_archived ...)

    if ( $XML_is_archived && !$ini_is_archived )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tFolder XML says archived and ini says not\n";
      $folder_error_count++;
    } ## end if ( $XML_is_archived ...)

    if ( $id && exists $folder_name_from{$id} )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tDuplicate folder ID: $id: $folder_name_from{$id}\n";
      $folder_error_count++;
      $folder_name_from{$id} = join ', ', $folder_name_from{$id},
        $this_name . '/' . $folder_name;
    } ## end if ( $id && exists $folder_name_from...)
    else
    {
      $folder_name_from{$id} = $this_name . '/' . $folder_name;
    }

    if (    exists $admin_file{'folder.bag'}
         && $admin_file{'folder.bag'}[0] == 0
         && scalar keys %folder_msg )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tfolder.bag is zero bytes and has *.msg files\n";
      $folder_error_count++;
    } ## end if ( exists $admin_file...)

    if ( exists $admin_file{'folder.bag'} && $admin_file{'folder.bag'}[0] )
    {
      if ( open my $BAG, '<', "$this_folder/folder.bag" )
      {
        while ( my $line = <$BAG> )
        {
          $bag_count++;
          my @fields = ( split /\xDE/, $line );
          my $basename = $fields[10];
          my ($name) = split /\./, lc $basename;

          if ( $fields[13] & 1 )
          {
            $bag_unread++;
            $bag_msg{$name} = [ $basename, 'unread' ];
          }
          else
          {
            $bag_msg{$name} = [ $basename, 'read' ];
          }
        } ## end while ( my $line = <$BAG>...)

        close $BAG;
      } ## end if ( open my $BAG, '<'...)
      else
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        print "\tCan't open $this_folder/folder.bag $!\n";
        $folder_error_count++;
      } ## end else [ if ( open my $BAG, '<'...)]
    } ## end if ( exists $admin_file...)

    if ( exists $admin_file{'folder.uct'}
         && $admin_file{'folder.uct'}[0] == 0 )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tfolder.uct is zero bytes\n";
      $folder_error_count++;
    } ## end if ( exists $admin_file...)
    else
    {
      if ( open my $UCT, '<', "$this_folder/folder.uct" )
      {
        $uct_unread = <$UCT> || 0;
        chomp $uct_unread;
        $uct_count = <$UCT> || 0;
        chomp $uct_count;
        close $UCT;
      } ## end if ( open my $UCT, '<'...)
      else
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        print "\tCan't open $this_folder/folder.uct $!\n";
        $folder_error_count++;
      } ## end else [ if ( open my $UCT, '<'...)]
    } ## end else [ if ( exists $admin_file...)]

    if ( is_valid_XML( "$this_folder/folderix.xml", \%admin_file ) )
    {
      $has_XML |= $IX_XML;

      if ( open my $XML, '<', "$this_folder/folderix.xml" )
      {
        local $/ = '</message>';

        while ( my $xml = <$XML> )
        {
          if ( $xml =~ m{basename="([^"]+).*?flags-msg="([^"]+)}ms )
          {
            $XML_ix_count++;
            my $basename = $1;
            my $flag     = $2;
            my ($name) = split /\./, lc $basename;

            if ( $flag & 1 )
            {
              $XML_ix_unread++;
              $XML_msg{$name} = [ $basename, 'unread' ];
            }
            else
            {
              $XML_msg{$name} = [ $basename, 'read' ];
            }
          } ## end if ( $xml =~ m{basename="([^"]+).*?flags-msg="([^"]+)}ms...)
        } ## end while ( my $xml = <$XML> ...)

        close $XML;
      } ## end if ( open my $XML, '<'...)
      else
      {
        warning( $this_folder, \$warned_file, $this_name,
                 $folder_name, \$warned_name );
        print "\tCan't open $this_folder/folderix.xml $!\n";
        $folder_error_count++;
      } ## end else [ if ( open my $XML, '<'...)]

    } ## end if ( is_valid_XML( "$this_folder/folderix.xml"...))

#print Data::Dumper->new([\%admin_file],[qw(admin)])->Indent(2)->Quotekeys(0)->Dump;
#print Data::Dumper->new([\%folder_msg],[qw(messages)])->Indent(2)->Quotekeys(0)->Dump;
#print Data::Dumper->new([\%bag_msg],[qw(bag)])->Indent(2)->Quotekeys(0)->Dump;
#print Data::Dumper->new([\%XML_msg],[qw(XML)])->Indent(2)->Quotekeys(0)->Dump;

    if ( $has_XML == $PR_XML
         && is_version_before( $pmmail_version, '(3.12.97.1926-27001)' ) )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tHas folderpr.xml but no folderix.xml\n";
      $folder_error_count++;
    } ## end if ( $has_XML == $PR_XML...)
    elsif ( $has_XML == $IX_XML )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tHas folderix.xml but no folderpr.xml\n";
      $folder_error_count++;
    } ## end elsif ( $has_XML == $IX_XML...)

    # folder against bag

  FOLDER_MESSAGE:
    foreach my $file ( sort keys %folder_msg )
    {
      $folder_msg_count++;
      my ( $size, $mtime, $full_name, $gz ) = @{ $folder_msg{$file} };

      if ( $is_archived && !defined $gz )
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        print "\tUncompressed Message $full_name in archived folder\n";
        $folder_error_count++;
      } ## end if ( $is_archived && !...)
      elsif ( !$is_archived && defined $gz )
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        print "\tCompressed Message $full_name in non archived folder\n";
        $folder_error_count++;
      } ## end elsif ( !$is_archived && ...)

      unless ($size)
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        print "\tMessage $full_name is zero bytes\n";
        $folder_error_count++;
      } ## end unless ($size)

      unless ( exists $bag_msg{$file} )
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
          = localtime($mtime);

        if ($max_errors)
        {
          printf
            "\tMessage $full_name (Size %d %d-%2.2d-%2.2d %2.2d:%2.2d) in folder not in bag\n (as $file.msg)",
            $size, $year + 1900, $mon + 1, $mday, $hour, $min
            if $message_error_count < $max_errors;
        } ## end if ($max_errors)
        else
        {
          printf
            "\tMessage $full_name (Size %d %d-%2.2d-%2.2d %2.2d:%2.2d) in folder not in bag (as $file.msg)\n",
            $size, $year + 1900, $mon + 1, $mday, $hour, $min;
        } ## end else [ if ($max_errors) ]

        $message_error_count++;
      } ## end unless ( exists $bag_msg{$file...})

      # folder against XML

      if ( ( $has_XML & $IX_XML ) && ( $XML_pr_count || $XML_ix_count ) )
      {
        next FOLDER_MESSAGE if exists $XML_msg{$file};
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        my ( $size, $mtime, $full_name, $gz ) = @{ $folder_msg{$file} };
        my ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst )
          = localtime($mtime);

        if ($max_errors)
        {
          printf
            "\tMessage $full_name (Size %d %d-%2.2d-%2.2d %2.2d:%2.2d) in folder not in XML (as $file.msg)\n",
            $size, $year + 1900, $mon + 1, $mday, $hour, $min
            if $message_error_count < $max_errors;
        } ## end if ($max_errors)
        else
        {
          printf
            "\tMessage $full_name (Size %d %d-%2.2d-%2.2d %2.2d:%2.2d) in folder not in XML (as $file.msg)\n",
            $size, $year + 1900, $mon + 1, $mday, $hour, $min;
        } ## end else [ if ($max_errors) ]

        $message_error_count++;
      } ## end if ( ( $has_XML & $IX_XML...))

    } ## end foreach my $file ( sort keys...)

    # bag against folder

  BAG_MESSAGE:
    foreach my $msg ( sort keys %bag_msg )
    {
      unless ( exists $folder_msg{$msg} )
      {
        my ( $full_name, $state ) = @{ $bag_msg{$msg} };
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );

        if ($max_errors)
        {
          print "\tMessage $full_name in bag not in folder\n"
            if $message_error_count < $max_errors;
        }
        else
        {
          print "\tMessage $full_name in bag not in folder\n";
        }

        $message_error_count++;
      } ## end unless ( exists $folder_msg...)

      # bag against XML

      if ( exists $XML_msg{$msg} )
      {
        next BAG_MESSAGE
          if $bag_msg{$msg}[1] eq $XML_msg{$msg}[1];  # check unread state
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );

        if ($max_errors)
        {
          print
            "\tMessage $bag_msg{$msg}[0] is $bag_msg{$msg}[1] in bag but $XML_msg{$msg}[1] in XML\n"
            if $message_error_count < $max_errors;
        } ## end if ($max_errors)
        else
        {
          print
            "\tMessage $bag_msg{$msg}[0] is $bag_msg{$msg}[1] in bag but $XML_msg{$msg}[1] in XML\n";
        }

        $message_error_count++;
      } ## end if ( exists $XML_msg{$msg...})
      elsif ( $has_XML & $IX_XML )
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        my ( $full_name, $state ) = @{ $bag_msg{$msg} };

        if ($max_errors)
        {
          print "\tMessage $full_name in bag not in XML\n"
            if $message_error_count < $max_errors;
        }
        else
        {
          print "\tMessage $full_name in bag not in XML\n";
        }

        $message_error_count++;
      } ## end elsif ( $has_XML & $IX_XML...)
    } ## end foreach my $msg ( sort keys...)

    # XML against folder

  XML_MESSAGE:
    foreach my $msg ( sort keys %XML_msg )
    {
      unless ( exists $folder_msg{$msg} )
      {
        print_warning_headers( $this_folder, \$warned_file, $this_name,
                               $folder_name, \$warned_name );
        my ( $full_name, $state ) = @{ $XML_msg{$msg} };

        if ($max_errors)
        {
          print "\tMessage $full_name in XML not in folder\n"
            if $message_error_count < $max_errors;
        }
        else
        {
          print "\tMessage $full_name in XML not in folder\n";
        }

        $message_error_count++;
      } ## end unless ( exists $folder_msg...)

      next XML_MESSAGE if exists $bag_msg{$msg};

      # XML against bag

      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      my ( $full_name, $state ) = @{ $XML_msg{$msg} };

      if ($max_errors)
      {
        print "\tMessage $full_name in XML not in bag\n"
          if $message_error_count < $max_errors;
      }
      else
      {
        print "\tMessage $full_name in XML not in bag\n";
      }

      $message_error_count++;
    } ## end foreach my $msg ( sort keys...)

    # counts etc.

    if ( $uct_unread != $bag_unread )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tUnread count mismatch: UCT: $uct_unread; BAG: $bag_unread\n";
      $folder_error_count++;
    } ## end if ( $uct_unread != $bag_unread...)

    if ( ( $has_XML == ( $PR_XML & $IX_XML ) )
      && ( $uct_unread != $XML_pr_unread || $XML_pr_unread != $XML_ix_unread )
      )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print
        "\tUnread count mismatch: UCT: $uct_unread; XML_PR: $XML_pr_unread; XML_IX: $XML_ix_unread\n";
      $folder_error_count++;
    } ## end if ( ( $has_XML == ( $PR_XML...)))
    elsif ( ( $has_XML & $PR_XML ) && ( $uct_unread != $XML_pr_unread ) )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print
        "\tUnread count mismatch: UCT: $uct_unread; XML_PR: $XML_pr_unread\n";
      $folder_error_count++;
    } ## end elsif ( ( $has_XML & $PR_XML...))
    elsif ( ( $has_XML & $IX_XML ) && ( $uct_unread != $XML_ix_unread ) )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print
        "\tUnread count mismatch: UCT: $uct_unread; XML_IX: $XML_ix_unread\n";
      $folder_error_count++;
    } ## end elsif ( ( $has_XML & $IX_XML...))

    my $folder_msg_count = keys %folder_msg;

    if ( (    $uct_count != $bag_count
           || $uct_count != $folder_msg_count
           || $bag_count != $folder_msg_count
         )
         || ( ( $has_XML & $PR_XML )
              && (    $uct_count != $XML_pr_count
                   || $XML_pr_count != $bag_count
                   || $XML_pr_count != $folder_msg_count )
         )
         || ( ( $has_XML & $IX_XML )
              && (    $uct_count != $XML_ix_count
                   || $XML_ix_count != $bag_count
                   || $XML_ix_count != $folder_msg_count )
         )
         || (    $has_XML == ( $PR_XML & $IX_XML )
              && $XML_pr_count != $XML_ix_count )
      )
    {
      print_warning_headers( $this_folder, \$warned_file, $this_name,
                             $folder_name, \$warned_name );
      print "\tMessage count mismatch: ";
      print "UCT: $uct_count; BAG: $bag_count; ";
      print "XML_PR: $XML_pr_count; " if $has_XML & $PR_XML;
      print "XML_IX: $XML_ix_count; " if $has_XML & $IX_XML;
      print "Actual: $folder_msg_count\n";
      $folder_error_count++;
    } ## end if ( ( $uct_count != $bag_count...))
  } ## end if ( $this_folder =~ m{\.fld$}i...)

  if ( $max_errors && $message_error_count > $max_errors )
  {
    printf "\n\t%d Message warnings were suppressd\n",
      $message_error_count - $max_errors;
  }

  # do the folders if any

  foreach my $folder (@folders)
  {
    if ( $folder_name && $folder_name ne '?' )
    {
      folder( $this_folder . '/' . $folder, $this_name . '/' . $folder_name );
    }
    else
    {
      folder( $this_folder . '/' . $folder, $this_name );
    }
  } ## end foreach my $folder (@folders...)

  $account_error_count += $folder_error_count + $message_error_count;
  return;
} ## end sub folder

sub is_valid_XML
{
  my $XML_file = shift;
  my $file_ref = shift;
  my $file_name;

  if ( $XML_file =~ m{\/([^\/]*)$} )
  {
    $file_name = $1;
  }

  return 0 unless exists $file_ref->{$file_name};
  my @tags;
  my $x   = -1;  # tag index
  my $tag = q{};

  if ( open my $XML_handle, '<', $XML_file )
  {
    <$XML_handle>;  # junk first line
    local $/ = '>';

    while ( my $line = <$XML_handle> )
    {
      if ( $line =~ m{<([^\s>/]+)} )  # <tag...  or <tag>
      {
        $tag = $1;
        $tags[ ++$x ] = $tag unless $line =~ m{/>$};
      }
      elsif ( $line =~ m{</([^>]+)>} )  # </tag>
      {
        if ( $1 ne $tags[$x] )
        {
          print "\nWARNING: XML file $XML_file suspect\n\tMisordered tags\n";
          print "\tFound <$tags[$x]> expected </$1>\n";
          $account_error_count++;
          close $XML_handle;
          return 0;
        } ## end if ( $1 ne $tags[$x] )

        $x--;
      } ## end elsif ( $line =~ m{</([^>]+)>}...)
    } ## end while ( my $line = <$XML_handle>...)
    close $XML_handle;
  } ## end if ( open my $XML_handle...)
  else
  {
    print "\nWARNING: Can't open $XML_file $!\n";
    $account_error_count++;
    return 0;
  } ## end else [ if ( open my $XML_handle...)]

  if ( $x != -1 )
  {
    print "\nWARNING: XML file $XML_file suspect\n\tMismatched tags\n";
    $account_error_count++;

    foreach my $i ( 0 .. $x )
    {
      print "\t" x $i;  # print $i tab chars
      print "<$tags[$i]...\n";
    }

    return 0;
  } ## end if ( $x != -1 )

  unless ($tag)
  {
    print "\nWARNING: XML file $XML_file suspect\n\tNo tags found\n";
    $account_error_count++;
    return 0;
  } ## end unless ($tag)

  return 1;  # show seems OK
} ## end sub is_valid_XML

sub commify
{
  local $_ = shift;
  1 while s{^(-?\d+)(\d{3})}{$1,$2};
  return $_;
}

sub print_warning_headers
{
  my $this_folder     = shift;
  my $warned_file_ref = shift;
  my $this_name       = shift;
  my $folder_name     = shift;
  my $warned_name_ref = shift;
  print "\nWARNING: $this_folder\n" unless $$warned_file_ref;
  $$warned_file_ref = 1;
  print "\t($this_name/$folder_name)\n" unless $$warned_name_ref;
  $$warned_name_ref = 1;
  return;
} ## end sub print_warning_headers

sub is_version_before
{
  my $this = shift;              # version dragged from log files
  my $old  = shift;              # version to compare against
  my @this = $this =~ m{(\d+)}g;
  my @old  = $old =~ m{(\d+)}g;

  return 0 if $#this > $#old;

  for ( my $x = 0; $x < @this; $x++ )
  {
    next if $this[$x] <= $old[$x];
    return 0;
  }

  return 1;
} ## end sub is_version_before
